from pydantic import BaseConfig, BaseModel


class WindowGeometry(BaseModel):
    """X11 window geometry abstraction."""

    x: int  # noqa: WPS111
    y: int  # noqa: WPS111
    width: int
    height: int
    border_width: int
    depth: int
    sequence_number: int

    class Config(BaseConfig):
        """Model config."""

        orm_mode = True


class ScreenGeometry(BaseModel):
    """Screen geometry parameters."""

    width: int
    height: int

    class Config(BaseConfig):
        """Model config."""

        orm_mode = True



from Xlib.protocol.display import Screen
from Xlib.xobject.drawable import Window

from s3wm_core.utils import get_screen_size
from s3wm_core.x_models import ScreenGeometry


class S3screen:
    """Screen abstraction for S3wm."""

    def __init__(self, screen: Screen):
        self.screen = screen

    @property
    def geom(self) -> ScreenGeometry:
        """
        Get screen parameters.

        :return: screen geometry.
        """
        width, height = get_screen_size(self.screen)
        return ScreenGeometry(
            width=width,
            height=height,
        )

    @property
    def root_window(self) -> Window:
        """
        Get root window.

        :return: root window.
        """
        return self.screen.root

    @property
    def width(self) -> int:
        """
        Screen width.

        Use geometry to support connected HDMI or DP.

        :return: Screen width in pixels.
        """
        return int(self.screen.width_in_pixels)

    @property
    def height(self) -> int:
        """
        Screen height.

        :return: Screen height in pixels.
        """
        return int(self.screen.height_in_pixels)



"""Default WM configuration."""
import sys
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path

from loguru import logger
from Xlib import X

from s3wm.layouts import DefaultTile
from s3wm_core import KeyCombination, kill_wm


def startup() -> None:
    """
    Actions to run at startup.

    Function called on startup.
    Actions defined here will be called at WindowManager startup.
    """
    logger.debug("Calling startup!")


# Default layout mode.
layout = DefaultTile
layout.gaps = 10

# Keyboard global combinations.
# All combinations for layoutManager
combinations = [
    KeyCombination(
        modifiers=KeyCombination.default_mod_key | X.ShiftMask,
        key="Return",
        action="xterm",
    ),
    KeyCombination(
        modifiers=KeyCombination.default_mod_key | X.ShiftMask,
        key="q",
        action=kill_wm,
    ),
    KeyCombination(
        modifiers=KeyCombination.default_mod_key,
        key="d",
        action="dmenu_run",
    ),
    *layout.get_keys(),
]


# Importing user_config bt absolute path.
module_name = "user_config"
conf_path = Path("~/.s3wm_conf.py").expanduser()

try:  # noqa: WPS229
    if not conf_path.exists():
        raise ImportError
    spec = spec_from_file_location(module_name, str(conf_path))
    module = module_from_spec(spec)
    if not spec.loader:
        raise ImportError
    sys.modules[module_name] = module
    spec.loader.exec_module(module)  # type: ignore
    from user_config import *  # noqa: F401, F403, WPS347, WPS433
except ImportError:
    logger.error("Can't import user config. Initialized with default.")



from typing import Any, Callable, Optional, Union

from Xlib import XK, X


class KeyCombination:
    """Used to add global key combinations."""

    default_mod_key = X.Mod4Mask

    def __init__(  # noqa: WPS234
        self,
        modifiers: int = X.NONE,
        key: Optional[Union[str, int]] = None,
        action: Optional[Union[Callable[..., Any], str]] = None,
    ) -> None:
        self.modifiers = modifiers  # Event masks such as Mod4Mask
        self.key = key  # Key of event
        if isinstance(key, str):
            self.key = XK.string_to_keysym(key)
        self.action = action  # Action to perform when event pressed

    def __repr__(self) -> str:
        return f"{{key:{self.key} mod:{self.modifiers}"



import re
import subprocess
from typing import Optional, Tuple

from loguru import logger
from Xlib.protocol.display import Screen
from Xlib.xobject.drawable import Window

from s3wm_core.x_models import WindowGeometry


def get_window_class(window: Window) -> str:
    """
    Get window wm_class.

    Fetch the WM_CLASS window property of the window WINDOW and return
    the class part of the property.  Return empty string if class is not
    retrieved.

    :param window: target window.
    :returns: window wm_class string.
    """
    try:
        cmd, window_cls = window.get_wm_class()
    except Exception as exc:
        logger.error(f"Can't get window class. Cause: {exc}")
        return ""
    if window_cls is not None:
        return str(window_cls)
    return ""


def get_window_geometry(window: Window) -> Optional[WindowGeometry]:
    """
    Obtain the geometry and attributes of the X11 window.

    :param window: target window.
    :returns: WindowGeomerty wrapper around X11 geometry type.
    """
    try:
        return WindowGeometry.from_orm(window.get_geometry())
    except Exception as exc:
        logger.exception(exc)
        logger.error(f"Can't get window geometry. Cause: {exc}")
        return None


def get_screen_size(screen: Screen) -> Tuple[int, int]:
    """
    Get current screen size.

    Return the dimension (WIDTH, HEIGHT) of the current screen as a
    tuple in pixels.  If xrandr command exists and either DP (DisplayPort)
    or HDMI output is active, return its dimensions instead of the current X11 display
    size.

    :param screen: current screen.
    :returns: Width and height of current screen.
    """
    width, height = screen.width_in_pixels, screen.height_in_pixels
    output = subprocess.getoutput("xrandr --current")
    # pick the last line including DP- or HDMI-
    match = re.search(r"(DP-?\d|HDMI-?\d) connected (\d+)x(\d+)", output)
    if match:
        width = int(match.group(2))
        height = int(match.group(3))
    logger.debug(f"get_screen_size -> w:{width} h:{height}")
    return width, height


def get_usable_screen_size(screen: Screen) -> Tuple[int, int]:
    """
    Get usable dimensions of the current screen.

    Return the dimension (WIDTH, HEIGHT) of the usable screen are
    (i.e., the area of the current screen excluding the are for displaying
    status monitor using, for example, xpymon.

    :param screen: current screen.
    :returns: Width and height of current screen.
    """
    width, height = get_screen_size(screen)
    logger.debug(f"get_usable_screen_size -> w:{width} h:{height}")
    return width, height



from typing import Callable

from s3wm.s3wm import S3WM


def change_tab(tab_index: int) -> Callable[[S3WM], None]:
    """
    Create function to change tab on layout.

    :param tab_index: index of a tab.
    :return: function to change tab on layout.
    """

    def change_layout_tab(wm: S3WM) -> None:
        """
        Actually change layout's tab.

        :param wm: WindowManager.
        """
        wm.layout.change_tab(tab_index)

    return change_layout_tab


def move_focus(prev: bool) -> Callable[[S3WM], None]:
    """
    Change currently focused window.

    This function moves focus backward or forward, depending on a "prev" parameter.

    :param prev: move focus on a previous window.
    :return: function to change focus.
    """

    def change_focus(wm: S3WM) -> None:
        """
        Actually change focus.

        :param wm: current window manager.
        """
        if prev:
            wm.layout.focus_prev()
        else:
            wm.layout.focus_next()

    return change_focus


def move_window_to_tab(tab_index: int) -> Callable[[S3WM], None]:
    """
    Generate function to move focused window.

    :param tab_index: tab on which to move window.
    :returns: function to move window.
    """

    def move_window(wm: S3WM) -> None:
        """
        Move currently focused window to another tab.

        :param wm: window manager.
        """
        wm.layout.move_focused_window(tab_index)

    return move_window


def kill_focused_window(wm: S3WM) -> None:
    """
    Kill window that user currently focused at.

    :param wm: window manager.
    """
    wm.layout.kill_focused_window()


def change_gaps(delta: int) -> Callable[[S3WM], None]:
    """
    Change gaps between windows.

    :param delta: delta to add to current gaps value.
    :returns: function to change gaps.
    """

    def gap_changer(wm: S3WM) -> None:
        """
        Actual function to update gaps.

        :param wm: window manager.
        """
        wm.layout.change_gap_value(delta)

    return gap_changer


def change_window_size(delta: int) -> Callable[[S3WM], None]:
    """
    Function to generate function to change window size.

    :param delta: size delta.
    :returns: function to change window size.
    """

    def size_changer(wm: S3WM) -> None:
        """
        Actually change size.

        :param wm: window manager.
        """
        wm.layout.change_main_window_size(delta)

    return size_changer


def move_window_forward(wm: S3WM) -> None:
    """
    Move focused window forward on current tab.

    :param wm: window manager.
    """
    wm.layout.move_window_forward()


def move_window_backward(wm: S3WM) -> None:
    """
    Move focused window backward on current tab.

    :param wm: window manager.
    """
    wm.layout.move_window_backward()



from typing import Any, Optional

from Xlib.X import CurrentTime, RevertToParent
from Xlib.xobject.drawable import Window

from s3wm_core.s3screen import S3screen
from s3wm_core.utils import get_window_geometry
from s3wm_core.x_models import WindowGeometry


class S3window(object):
    """Main window abstraction for S3WM."""

    def __init__(
        self,
        window: Window,
        screen: S3screen,
        parent: Optional[Window] = None,
    ):
        self.parent = parent
        self.window = window
        self.screen = screen

    @property
    def id(self) -> int:
        """
        X11 resource ID allocated for this window.

        :return: unique window ID.
        """
        return int(self.window.id)

    @property
    def geom(self) -> Optional[WindowGeometry]:
        """
        Get window geometry.

        :return: Window geometry
        """
        return get_window_geometry(self.window)

    def map(self) -> None:
        """Maps window in X11."""
        self.window.map()

    def unmap(self) -> None:
        """Unmap window in X11."""
        self.window.unmap()

    def focus(self) -> None:
        """Set focus to window."""
        self.window.warp_pointer(15, 15)  # noqa: WPS432
        self.window.set_input_focus(
            RevertToParent,
            CurrentTime,
        )

    def resize(self, width: int, height: int, percents: bool = False) -> None:
        """
        Resize window.

        If percents mode is on, width and height are treated as a percents,
        relative to the current screen geometry.

        :param width: new width
        :param height: new height
        :param percents: per cent mode is on or off.
        """
        win_width = width
        win_height = height
        if percents:
            screen_geom = self.screen.geom
            win_width = (screen_geom.width * width) // 100
            win_height = (screen_geom.height * height) // 100
        self.window.configure(
            width=win_width,
            height=win_height,
        )

    def move(self, x: int, y: int) -> None:  # noqa: WPS111
        """
        Move window on the screen.

        :param x: top left corner x coordinate.
        :param y: top left corner y coordinate.
        """
        self.window.configure(
            x=x,
            y=y,
        )

    def destroy(self) -> None:
        """Kill window from X11."""
        self.window.destroy()

    def __str__(self) -> str:
        return f"<S3Window {self.id}>"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, S3window):
            return False
        return self.id == other.id



import logging
from argparse import ArgumentParser, Namespace
from enum import Enum
from importlib.metadata import version
from sys import stdout

from loguru import logger

from s3wm.s3wm import S3WM


class Loglevel(Enum):
    """Different log levels."""

    DEBUG = logging.DEBUG
    INFO = logging.INFO  # noqa: WPS110
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    def __str__(self) -> str:
        return str(self.name)

    @staticmethod
    def from_string(level: str) -> "Loglevel":
        """
        Translate string to enum.

        :param level: level's string representation.
        :return: parsed enum if any.
        :raises ValueError: if can't find a value.
        """
        try:
            return Loglevel[level]
        except KeyError:
            raise ValueError("Unknown level")


def parse_arguments() -> Namespace:
    """
    Parse CLI arguments passed to s3wm.

    :return: namespace with parsed args.
    """
    parser = ArgumentParser()
    parser.add_argument(
        "--version",
        "-v",
        action="store_true",
        dest="version",
    )
    parser.add_argument(
        "--log-level",
        "-l",
        dest="log_level",
        type=Loglevel.from_string,
        choices=list(Loglevel),
        default=Loglevel.INFO,
    )
    return parser.parse_args()


def main() -> None:
    """Function to run the thing."""
    args = parse_arguments()
    if args.version:
        s3wm_version = version("s3wm")
        print(f"S3WM version: {s3wm_version}")  # noqa: WPS421
        return
    logger.remove()
    logger.add(stdout, level=args.log_level.value)
    wm = S3WM()
    wm.run()



from collections import defaultdict
from typing import DefaultDict, List

from Xlib.X import ShiftMask

from s3wm.layouts.default_tile.key_bindings import (
    change_gaps,
    change_tab,
    change_window_size,
    kill_focused_window,
    move_focus,
    move_window_backward,
    move_window_forward,
    move_window_to_tab,
)
from s3wm.layouts.default_tile.tab import Tab
from s3wm.s3wm import S3WM
from s3wm_core.key_combination import KeyCombination
from s3wm_core.layout_base import AbstractLayoutManager
from s3wm_core.s3window import S3window


class DefaultTile(AbstractLayoutManager):
    """Default tile manager."""

    gaps: int = 0
    tab_class = Tab

    def __init__(self, wm: S3WM) -> None:
        """
        Initialize Layout.

        :param wm: S3WM instance.
        """
        super().__init__(wm)
        self.tab_class.gaps = self.gaps
        self.tabs: DefaultDict[int, Tab] = defaultdict(self.tab_class)
        self.current_tab = 0
        self.tabs[self.current_tab].focus()

    def add_window(self, window: S3window) -> None:
        """
        Add window to layout manager and adjust size and position.

        :param window: new window.
        """
        self.tabs[self.current_tab].add_window(window)

    def remove_window(self, window: S3window) -> None:
        """
        Remove window from current layout.

        :param window: removed window.
        """
        self.tabs[self.current_tab].remove_window(window)

    def change_tab(self, tab_number: int) -> None:
        """
        Select tab to show on screen.

        :param tab_number: tab index
        """
        if tab_number == self.current_tab:
            return
        self.tabs[self.current_tab].lose_focus()
        self.current_tab = tab_number
        self.tabs[self.current_tab].focus()

    def focus_next(self) -> None:
        """Focus next window on the current tab."""
        self.tabs[self.current_tab].focus_next()

    def focus_prev(self) -> None:
        """Focus previous window on the current tab."""
        self.tabs[self.current_tab].focus_prev()

    def move_focused_window(self, tab_index: int) -> None:
        """
        Move currently focused window to another tab.

        :param tab_index: to which tab do we need to move.
        """
        window = self.tabs[self.current_tab].pop_focused_window()
        if window:
            self.tabs[tab_index].add_window(window)

    def kill_focused_window(self) -> None:
        """Kill focused window."""
        window = self.tabs[self.current_tab].pop_focused_window()
        if window:
            window.destroy()

    def change_gap_value(self, gap_delta: int) -> None:
        """
        Update gap value for tabs.

        :param gap_delta: delta to current gap value.
        """
        self.tab_class.gaps = min(self.tab_class.gaps + gap_delta, 100)
        self.tab_class.gaps = max(0, self.tab_class.gaps)
        self.tabs[self.current_tab].update_layout()

    def change_main_window_size(self, delta: int) -> None:
        """
        Change main window size.

        :param delta: size delta.
        """
        self.tabs[self.current_tab].change_main_window_size(delta)

    def move_window_forward(self) -> None:
        """Move currently focused window forward."""
        self.tabs[self.current_tab].move_window_forward()

    def move_window_backward(self) -> None:
        """Move currently focused window backward."""
        self.tabs[self.current_tab].move_window_backward()

    @classmethod
    def get_keys(cls) -> List[KeyCombination]:
        """Get Keys specific to your layout.

        This method need to return keyCombinations to
        manipulate your layout.

        :return: List of key combinations
        """
        keys = super().get_keys()
        for index in range(1, 9 + 1):
            keys.append(
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key,
                    key=str(index),
                    action=change_tab(index - 1),
                ),
            )
            keys.append(
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key | ShiftMask,
                    key=str(index),
                    action=move_window_to_tab(index - 1),
                ),
            )

        keys.extend(
            [
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key,
                    key="j",
                    action=move_focus(prev=False),
                ),
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key,
                    key="k",
                    action=move_focus(prev=True),
                ),
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key | ShiftMask,
                    key="j",
                    action=move_window_backward,
                ),
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key | ShiftMask,
                    key="k",
                    action=move_window_forward,
                ),
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key | ShiftMask,
                    key="c",
                    action=kill_focused_window,
                ),
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key,
                    key="minus",
                    action=change_gaps(-2),
                ),
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key,
                    key="plus",
                    action=change_gaps(2),
                ),
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key,
                    key="h",
                    action=change_window_size(-5),
                ),
                KeyCombination(
                    modifiers=KeyCombination.default_mod_key,
                    key="l",
                    action=change_window_size(5),
                ),
            ],
        )

        return keys



"""
Core functions and modules for S3WM.

It's usable for creating new layout managers and stuff.
"""
from s3wm_core.key_combination import KeyCombination
from s3wm_core.keymap import kill_wm
from s3wm_core.layout_base import AbstractLayoutManager
from s3wm_core.s3screen import S3screen
from s3wm_core.s3window import S3window
from s3wm_core.x_models import ScreenGeometry, WindowGeometry

__all__ = [
    "kill_wm",
    "S3window",
    "S3screen",
    "KeyCombination",
    "WindowGeometry",
    "ScreenGeometry",
    "AbstractLayoutManager",
]



from subprocess import Popen

from frozendict import frozendict
from loguru import logger
from Xlib import X
from Xlib.display import Display
from Xlib.protocol.event import (
    DestroyNotify,
    EnterNotify,
    KeyPress,
    LeaveNotify,
    MapRequest,
    UnmapNotify,
)
from Xlib.Xcursorfont import left_ptr

from s3wm_core.keymap import get_key_action, init_keymap
from s3wm_core.s3screen import S3screen
from s3wm_core.s3window import S3window

EVENT_HANDLER_MAP = frozendict(
    {
        X.KeyPress: "handle_keypress",
        X.ButtonPress: None,
        X.MotionNotify: None,
        X.ButtonRelease: None,
        X.MapRequest: "handle_map",
        X.ConfigureRequest: None,
        X.UnmapNotify: "handle_unmap",
        X.EnterNotify: "handle_focus_in",
        X.LeaveNotify: "handle_focus_out",
        X.DestroyNotify: "handle_destroy",
        X.MapNotify: None,
    },
)


class S3WM:
    """
    Main window manager class.

    This class is a bunch of X11 callbacks and setup functions.
    Main
    """

    def __init__(self) -> None:
        """
        Initialize S3WM.

        Initialization process includes
        establishing connection to the display
        and initialization of a chosen layout.
        """
        from s3wm_core import wm_config  # noqa: WPS433

        self.display = Display()
        self.config = wm_config
        self.layout = wm_config.layout(self)
        font = self.display.open_font("cursor")
        cursor = font.create_glyph_cursor(  # noqa: WPS317
            font,
            left_ptr,
            left_ptr + 1,
            (0, 0, 0),
            (65535, 65535, 65535),
        )
        self.display.screen().root.change_attributes(cursor=cursor)

    def handle_map(self, map_event: MapRequest) -> None:
        """
        Map window to layout.

        Map events called when new window is created.
        At this point we need to calculate the size of
        window and place it nicely at your workspace.

        :param map_event: X11 event for mapping
        """
        logger.debug("Mapping window")
        window = S3window(map_event.window, S3screen(self.display.screen()))
        window.map()
        self.layout.add_window(window)
        mask = X.EnterWindowMask | X.LeaveWindowMask
        window.window.change_attributes(event_mask=mask)

    def handle_keypress(self, key_event: KeyPress) -> None:
        """
        Run command on keypress.

        This event called when user using defined key combination and run task
        assigned to this combination.

        :param key_event: Key combination.
        """
        action = get_key_action(key_event)
        if not action:
            return
        if callable(action):
            logger.debug("Found python function")
            action(self)
        else:
            logger.debug(f"Running os command: '{action}'")
            Popen(action, shell=True)

    def handle_destroy(self, destroy_event: DestroyNotify) -> None:
        """
        Called to destroy window.

        This function will be triggered when window is destroyed.
        :param destroy_event: X11 event.
        """
        self.layout.remove_window(
            S3window(destroy_event.window, S3screen(self.display.screen())),
        )

    def handle_unmap(self, unmap_event: UnmapNotify) -> None:
        """
        Called to unmap window and remove it from the screen.

        This function will be triggered when window is unmapped.
        :param unmap_event: X11 event.
        """
        self.layout.remove_window(
            S3window(unmap_event.window, S3screen(self.display.screen())),
        )

    def handle_focus_in(self, enter_event: EnterNotify) -> None:
        """
        Called when window gets focus.

        :param enter_event: X11 event.
        """
        self.layout.focus_in(
            S3window(enter_event.window, S3screen(self.display.screen())),
        )

    def handle_focus_out(self, leave_event: LeaveNotify) -> None:
        """
        Called when window lose focus.

        :param leave_event: X11 event.
        """
        self.layout.focus_out(
            S3window(leave_event.window, S3screen(self.display.screen())),
        )

    def catch_events(self) -> None:
        """
        Setup event catching.

        Configure the root window to receive all events
        needed for managing windows.
        """
        mask = (
            X.SubstructureRedirectMask
            | X.StructureNotifyMask
            | X.UnmapNotify
            | X.EnterWindowMask
            | X.LeaveWindowMask
            | X.FocusChangeMask
        )
        self.display.screen().root.change_attributes(event_mask=mask)

    def setup_root(self) -> None:
        """
        Setting up main window.

        This function will set all needed variables to
        this WindowManager.
        """
        logger.debug("Setting up root window")
        root_window = self.display.screen().root
        wm_name = self.display.intern_atom("_NET_WM_NAME")
        utf_string = self.display.intern_atom("UTF8_STRING")
        # That thing needed only for Java applications.
        # Because Java can't handle custom window managers.
        root_window.change_text_property(wm_name, utf_string, "LG3D")
        self.display.sync()

    def handle_next_event(self) -> None:  # noqa: C901, WPS231
        """
        Request next event from X11 and handle it.

        :raises KeyboardInterrupt: if something has enterrupted the main process.
        """
        event = self.display.next_event()
        logger.debug(f"Received event: {event.__class__}")
        if event.type in EVENT_HANDLER_MAP:
            handler_name = EVENT_HANDLER_MAP.get(event.type)
            if not handler_name:
                return
            event_handler = getattr(self, handler_name)
            if event_handler:
                logger.debug(f"Found event_handler: {handler_name}")
                try:
                    event_handler(event)
                except KeyboardInterrupt:  # noqa: WPS329
                    raise
                except Exception as exc:
                    logger.exception(exc)
                logger.debug("event handled")

    def reload_windows(self) -> None:
        """Query root window for children and render them if we can."""
        response = self.display.screen().root.query_tree()
        for window in response.children:
            attrs = window.get_attributes()
            if attrs.map_state == X.IsViewable:
                managed_window = S3window(
                    window,
                    S3screen(self.display.screen()),
                )
                managed_window.map()
                self.layout.add_window(managed_window)
                mask = X.EnterWindowMask | X.LeaveWindowMask
                window.change_attributes(event_mask=mask)

    def run(self) -> None:
        """Runs window manager."""
        display = self.display
        init_keymap(display)
        startup = getattr(
            self.config,
            "startup",
            lambda: logger.debug("No startup actions found"),
        )
        self.catch_events()
        self.setup_root()
        startup()
        self.reload_windows()
        while True:  # noqa: WPS457
            self.handle_next_event()



from typing import Any, Callable, Dict, Optional, Tuple, Union

from loguru import logger
from Xlib import X
from Xlib.display import Display
from Xlib.protocol.event import KeyPress

from s3wm_core.key_combination import KeyCombination

keycode_mapping: Dict[Tuple[int, int], KeyCombination] = {}


def kill_wm(wm: Any) -> None:
    """Close WM process.

    :param wm: an S3WM instance. (Used Any to avoid circular deps)
    """
    wm.display.close()
    exit(0)  # noqa: WPS421


def init_keymap(display: Display) -> None:
    """
    Sends requests to the X Server to listen for specific key events.

    When the key combination is pressed the application will get the event about it.

    :param display: Used to manipulate keysym to keycode transitions.
    """
    from s3wm_core.wm_config import combinations  # noqa: WPS433

    for button in {1, 3}:  # noqa: WPS335
        display.screen().root.grab_button(
            button=button,
            modifiers=X.Mod1Mask,
            owner_events=True,
            event_mask=X.ButtonPressMask,
            pointer_mode=X.GrabModeAsync,
            keyboard_mode=X.GrabModeAsync,
            confine_to=X.NONE,
            cursor=X.NONE,
        )

    # Registering all keybindings to X11.
    for combination in combinations:
        # Getting code for key on current keyboard.
        codes = {code for code, index in display.keysym_to_keycodes(combination.key)}

        for code in codes:
            logger.debug(combination)
            display.screen().root.grab_key(
                code,
                combination.modifiers,
                1,
                X.GrabModeAsync,
                X.GrabModeAsync,
            )
            keycode_mapping[(code, combination.modifiers)] = combination


def get_key_action(  # noqa: WPS234
    key_event: KeyPress,
) -> Optional[Union[Callable[..., Any], str]]:
    """
    Function to get the action defined in `combinations` by keypress event.

    :param key_event: event generated by X11.
    :return: Action associated with keypress event.
    """
    key = key_event.detail
    modifiers = key_event.state
    combination = keycode_mapping.get((key, modifiers))
    if combination:
        return combination.action
    return None



"""Default tile layout. Similar to DWM's default."""
from s3wm.layouts.default_tile.key_bindings import (
    change_gaps,
    change_tab,
    kill_focused_window,
    move_focus,
    move_window_to_tab,
)
from s3wm.layouts.default_tile.layout import DefaultTile

__all__ = [
    "DefaultTile",
    "kill_focused_window",
    "move_focus",
    "move_window_to_tab",
    "change_tab",
    "change_gaps",
]



from abc import ABC, abstractmethod
from typing import Any, List

from s3wm_core.key_combination import KeyCombination
from s3wm_core.s3window import S3window


class AbstractLayoutManager(ABC):
    """Base class for Layout managers."""

    def __init__(self, _wm: Any) -> None:
        """Do whatever you want with data."""

    @abstractmethod
    def add_window(self, window: S3window) -> None:
        """
        Add window to Layout manager.

        This method must put window under control of current LayoutManager.
        :param window: added window
        """

    @abstractmethod
    def remove_window(self, window: S3window) -> None:
        """
        Remove window from LayoutManager.

        :param window: window that was removed.
        """

    @classmethod
    def get_keys(cls) -> List[KeyCombination]:
        """Get Keys specific to your layout.

        This method need to return keyCombinations to
        manipulate your layout.

        :return: List of key combinations
        """
        return []

    def focus_in(self, window: S3window) -> None:
        """
        Notify about focusing on some specific window.

        This function is used to be called when pointer is moved to some window.

        :param window: window that pointer focusing at.
        """

    def focus_out(self, window: S3window) -> None:
        """
        Notify about event when pointer has stopped focusing on some window.

        This function is used to be called when pointer has left some window.

        :param window: window that lost focus.
        """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['s3wm', 's3wm.layouts', 's3wm.layouts.default_tile', 's3wm.tests', 's3wm_core']

package_data = \
{'': ['*']}

install_requires = \
['frozendict>=1.2,<2.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.7.3,<2.0.0',
 'python-xlib>=0.29,<0.30']

entry_points = \
{'console_scripts': ['s3wm = s3wm.main:main']}

setup_kwargs = {
    'name': 's3wm',
    'version': '0.2.0a0',
    'description': 'Simple Window manager.',
    'long_description': "![GitHub Workflow Status](https://img.shields.io/github/workflow/status/s3rius/s3wm/Release%20s3wm?style=for-the-badge)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/s3wm?style=for-the-badge)](https://pypi.org/project/s3wm)\n[![PyPI](https://img.shields.io/pypi/v/s3wm?style=for-the-badge)](https://pypi.org/project/s3wm)\n\n# S3WM\n\n🛠️⚙️This is a WIP project. Don't beleive in README ⚙️🛠️\n\n\nThis project is a yet another `Window manager`.\n\nMain Idea behind this project is `modularity` and\nwindow manager `configuration in Python`.\n\nYou can even `bind a python function` to some key combination.\n\n## Configuration\nMain configuration file must be located in `$HOME/.s3wm_conf.py`.\nS3WM configuration examples can be found in examples folder.\n\n## How to Run/test\n\n```bash\n# start nested X11 session with Xephyr\nXephyr :1 +xinerama -screen 1280x720 -reset -terminate &\n# Run it with proper DISPLAY ENV\nDISPLAY=:1.0 s3wm\n# To grab or release host keys press `Ctrl` + `Shift`\n```\n",
    'author': 'Pavel Kirilin',
    'author_email': 'win10@list.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List, Optional

from loguru import logger

from s3wm_core.s3window import S3window


class Tab:
    """Tab abstraction. This abstraction used in tiles to organize windows."""

    gaps: int = 0

    def __init__(self) -> None:
        self.windows: List[S3window] = []
        self.focused_window: Optional[S3window] = None
        self.main_window_size = 50

    def focus(self) -> None:
        """Show all windows from tab and focus on the last one."""
        for window in self.windows:
            window.map()
        self.update_layout()
        self.focused_window = None
        if self.windows:
            self.windows[-1].focus()

    def lose_focus(self) -> None:
        """Hide all windows from the screen."""
        for window in self.windows:
            window.unmap()

    def focused_index(self) -> int:
        """
        Get array index of currently focused window.

        :return: index.
        """
        for index, window in enumerate(self.windows):
            if window == self.focused_window:
                return index
        return 0

    def focus_prev(self) -> None:
        """Focus on a previous window in array."""
        if not self.windows:
            return
        index = self.focused_index()
        index += 1
        if index == len(self.windows):
            index = 0
        self.focused_window = self.windows[index]
        self.focused_window.focus()

    def focus_next(self) -> None:
        """Focus on a next window in windows stack."""
        if not self.windows:
            return
        index = self.focused_index()
        self.focused_window = self.windows[index - 1]
        self.focused_window.focus()

    def pop_focused_window(self) -> Optional[S3window]:
        """
        Pop focused window from tab.

        :return: previously focused window.
        """
        if not self.windows:
            return None
        index = self.focused_index()
        target_window = self.windows.pop(index)
        target_window.unmap()
        self.focused_window = None
        self.update_layout()
        return target_window

    def add_window(self, window: S3window) -> None:
        """
        Add new window to tab.

        :param window: new window.
        """
        self.windows.append(window)
        self.focused_window = window
        window.focus()
        self.update_layout()

    def remove_window(self, window: S3window) -> None:
        """
        Remove window from tab.

        :param window: removed window.
        """
        # Trying to remove window from windows list.
        for index, tab_window in enumerate(self.windows):
            if tab_window == window:
                # If we have such window we can unmap it.
                self.windows.pop(index)
                window.unmap()
        self.focused_window = None
        # If we have other windows on our tab we focus on the last one.
        if self.windows:
            self.windows[-1].focus()
            self.focused_window = self.windows[-1]
        self.update_layout()

    def change_main_window_size(self, diff: int) -> None:
        """
        Change width of a main window.

        :param diff: delta to current size.
        """
        if not self.windows or len(self.windows) == 1:
            return
        self.main_window_size = min(self.main_window_size + diff, 90)  # noqa: WPS432
        self.main_window_size = max(10, self.main_window_size)
        self.update_layout()

    def move_window_forward(self) -> None:
        """Move focused window backward in stack."""
        if not self.windows:
            return
        index = self.focused_index()
        second_index = index + 1
        if second_index >= len(self.windows):
            second_index = 0
        tmp = self.windows[index]
        self.windows[index] = self.windows[second_index]
        self.windows[second_index] = tmp
        self.update_layout()

    def move_window_backward(self) -> None:
        """Move focused window forward in stack."""
        if not self.windows:
            return
        index = self.focused_index()
        tmp = self.windows[index]
        self.windows[index] = self.windows[index - 1]
        self.windows[index - 1] = tmp
        self.update_layout()

    def update_layout(self) -> None:  # noqa: C901, WPS210, WPS231, WPS213
        """Place all windows on layout nicely."""
        logger.debug("Updating layout")
        logger.info(self.gaps)
        if not self.windows:
            return

        main_window_width = self.main_window_size
        main_width_gaps = 1

        if len(self.windows) == 1:
            main_window_width = 100
            main_width_gaps = 2

        main_window = self.windows[-1]
        main_window.resize(width=main_window_width, height=100, percents=True)
        main_window.move(
            x=self.gaps,
            y=self.gaps,
        )
        main_geom = main_window.geom
        if not main_geom:
            return
        main_window.resize(
            width=main_geom.width - (self.gaps * main_width_gaps),
            height=main_geom.height - (self.gaps * 2),
        )
        if len(self.windows) == 1:
            return
        main_geom = main_window.geom
        if not main_geom:
            return
        height_percent = 100 // (len(self.windows) - 1)
        windows_x = main_geom.x + main_geom.width + self.gaps
        windows_y = 0
        for window in reversed(self.windows):
            if window == main_window:
                continue
            window.resize(
                width=100 - self.main_window_size,  # : WPS432
                height=height_percent,
                percents=True,
            )
            win_geom = window.geom
            if not win_geom:
                continue
            window.resize(
                width=win_geom.width - (self.gaps * 2),
                height=win_geom.height - (self.gaps * 2),
            )
            window.move(x=windows_x, y=windows_y + self.gaps)
            windows_y += win_geom.height
